import numpy as np
from PyQt5.QtWidgets import QApplication, QMessageBox, QFrame, QSizePolicy, QWidget, QLineEdit, QListWidget, QVBoxLayout, QComboBox, \
	QPushButton, QLabel, QHBoxLayout, QCheckBox, QFileDialog
from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QDoubleValidator, QIntValidator

from celldetective.gui import Styles
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT
import matplotlib.pyplot as plt
import celldetective.extra_properties as extra_properties
from inspect import getmembers, isfunction
from celldetective.filters import *
from os import sep


class QuickSliderLayout(QHBoxLayout):
	
	"""
	A layout class that combines a QLabel and a QSlider in a horizontal box layout.

	This layout provides a convenient way to include a slider with an optional label and configurable 
	parameters such as range, precision, and step size. It allows for both integer and decimal step values, 
	making it versatile for different types of input adjustments.

	Parameters
	----------
	label : str, optional
		The label to be displayed next to the slider (default is None).
	slider : QSlider
		The slider widget to be added to the layout.
	layout_ratio : tuple of float, optional
		Defines the width ratio between the label and the slider in the layout. The first element is the 
		ratio for the label, and the second is for the slider (default is (0.25, 0.75)).
	slider_initial_value : int or float, optional
		The initial value to set for the slider (default is 1).
	slider_range : tuple of int or float, optional
		A tuple specifying the minimum and maximum values for the slider (default is (0, 1)).
	slider_tooltip : str, optional
		Tooltip text to display when hovering over the slider (default is None).
	decimal_option : bool, optional
		If True, the slider allows decimal values with a specified precision (default is True).
	precision : float, optional
		The step size for the slider when `decimal_option` is enabled (default is 1.0E-03).
	
	Attributes
	----------
	qlabel : QLabel
		The label widget that displays the provided label text (only if `label` is provided).
	slider : QSlider
		The slider widget that allows the user to select a value.
	"""

	def __init__(self, label=None, slider=None, layout_ratio=(0.25,0.75), slider_initial_value=1, slider_range=(0,1), slider_tooltip=None, decimal_option=True, precision=1.0E-03, *args):
		super().__init__(*args)

		if label is not None and isinstance(label,str):
			self.qlabel = QLabel(label)
			self.addWidget(self.qlabel, int(100*layout_ratio[0]))

		self.slider = slider
		self.slider.setOrientation(1)
		if decimal_option:
			self.slider.setSingleStep(precision)
			self.slider.setTickInterval(precision)
		else:
			self.slider.setSingleStep(1)
			self.slider.setTickInterval(1)

		self.slider.setRange(*slider_range)
		self.slider.setValue(slider_initial_value)
		if isinstance(slider_tooltip,str):
			self.slider.setToolTip(slider_tooltip)

		self.addWidget(self.slider, int(100*layout_ratio[1]))


def center_window(window):

	"""
	Centers the given window in the middle of the screen.

	This function calculates the current screen's geometry and moves the 
	specified window to the center of the screen. It works by retrieving the 
	frame geometry of the window, identifying the screen where the cursor is 
	currently located, and adjusting the window's position to be centrally 
	aligned on that screen.

	Parameters
	----------
	window : QMainWindow or QWidget
		The window or widget to be centered on the screen.
	"""

	frameGm = window.frameGeometry()
	screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
	centerPoint = QApplication.desktop().screenGeometry(screen).center()
	frameGm.moveCenter(centerPoint)
	window.move(frameGm.topLeft())

class ExportPlotBtn(QPushButton, Styles):
	
	"""
	A custom QPushButton widget for exporting a matplotlib figure.

	This class combines a QPushButton with functionality to export a given matplotlib 
	figure (`fig`) to an image file. The button includes an icon and a tooltip for easy 
	user interaction. When clicked, a file dialog is opened allowing the user to specify 
	the location and file format to save the plot.

	Parameters
	----------
	fig : matplotlib.figure.Figure
		The matplotlib figure object to be exported.
	export_dir : str, optional
		The default directory where the file will be saved. If not provided, the current 
		working directory will be used.
	*args : tuple
		Additional positional arguments passed to the parent `QPushButton` constructor.
	**kwargs : dict
		Additional keyword arguments passed to the parent `QPushButton` constructor.

	Attributes
	----------
	fig : matplotlib.figure.Figure
		The figure that will be saved when the button is clicked.
	export_dir : str or None
		The default directory where the file dialog will initially point when saving the image.

	Methods
	-------
	save_plot():
		Opens a file dialog to choose the file name and location for saving the figure.
		The figure is then saved in the specified format and location.
	"""

	def __init__(self, fig, export_dir=None, *args, **kwargs):

		super().__init__()

		self.export_dir = export_dir
		self.fig = fig

		self.setText('')
		self.setIcon(icon(MDI6.content_save,color="black"))
		self.setStyleSheet(self.button_select_all)
		self.setToolTip('Export figure.')
		self.setIconSize(QSize(20, 20))
		self.clicked.connect(self.save_plot)
	
	def save_plot(self):

		"""
		Opens a file dialog for the user to specify the location and name to save the plot.

		If the user selects a file, the figure is saved with tight layout and 300 DPI resolution.
		Supported formats include PNG, JPG, SVG, and XPM.
		"""

		if self.export_dir is not None:
			guess_dir = self.export_dir+sep+'plot.png'
		else:
			guess_dir = 'plot.png'
		fileName, _ = QFileDialog.getSaveFileName(self, 
			"Save Image", guess_dir, "Images (*.png *.xpm *.jpg *.svg)") #, options=options
		if fileName:
			self.fig.tight_layout()
			self.fig.savefig(fileName, bbox_inches='tight', dpi=300)


class QHSeperationLine(QFrame):
	'''
	a horizontal seperation line\n
	'''

	def __init__(self):
		super().__init__()
		self.setMinimumWidth(1)
		self.setFixedHeight(20)
		self.setFrameShape(QFrame.HLine)
		self.setFrameShadow(QFrame.Sunken)
		self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)


class FeatureChoice(QWidget):

	def __init__(self, parent_window):
		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Add feature")
		# Create the QComboBox and add some items
		self.combo_box = QComboBox(self)
		center_window(self)

		standard_measurements = ["area",
								 "area_bbox",
								 "area_convex",
								 "area_filled",
								 "major_axis_length",
								 "minor_axis_length",
								 "eccentricity",
								 "equivalent_diameter_area",
								 "euler_number",
								 "extent",
								 "feret_diameter_max",
								 "orientation",
								 "perimeter",
								 "perimeter_crofton",
								 "solidity",
								 "intensity_mean",
								 "intensity_max",
								 "intensity_min",
								 ]

		extra_props = getmembers(extra_properties, isfunction)
		extra_props = [extra_props[i][0] for i in range(len(extra_props))]
		if len(extra_props) > 0:
			standard_measurements.extend(extra_props)

		self.combo_box.addItems(standard_measurements)

		self.add_btn = QPushButton("Add")
		self.add_btn.clicked.connect(self.add_current_feature)

		# Create the layout
		layout = QVBoxLayout(self)
		layout.addWidget(self.combo_box)
		layout.addWidget(self.add_btn)

	def add_current_feature(self):
		filtername = self.combo_box.currentText()
		self.parent_window.list_widget.addItems([filtername])
		self.close()


class FilterChoice(QWidget):

	def __init__(self, parent_window):

		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Add filter")
		# Create the QComboBox and add some items
		center_window(self)

		self.default_params = {
			'gauss_filter': {'sigma': 2},
			'median_filter': {'size': 4},
			'maximum_filter': {'size': 4},
			'minimum_filter': {'size': 4},
			'percentile_filter': {'percentile': 99, 'size': 4},
			'variance_filter': {'size': 4},
			'std_filter': {'size': 4},
			'laplace_filter': None,
			'abs_filter': None,
			'ln_filter': None,
			'subtract_filter': {'value': 1},
			'dog_filter': {'sigma_low': 0.8, 'sigma_high': 1.6},
			'log_filter': {'sigma': 2},
			'tophat_filter': {'size': 4, 'connectivity': 4},
			'otsu_filter': None,
			'local_filter': {'block_size': 73, 'method': 'mean', 'offset': 0},
			'niblack_filter': {'window_size': 15, 'k': 0.2},
			# 'sauvola_filter': {'window_size': 15, 'k': 0.2}
		}

		layout = QVBoxLayout(self)
		self.combo_box = QComboBox(self)
		self.combo_box.addItems(list(self.default_params.keys()))
		self.combo_box.currentTextChanged.connect(self.update_arguments)
		layout.addWidget(self.combo_box)

		self.arguments_le = [QLineEdit() for i in range(3)]
		self.arguments_labels = [QLabel('') for i in range(3)]
		for i in range(2):
			hbox = QHBoxLayout()
			hbox.addWidget(self.arguments_labels[i], 20)
			hbox.addWidget(self.arguments_le[i], 80)
			layout.addLayout(hbox)

		self.add_btn = QPushButton("Add")
		self.add_btn.clicked.connect(self.add_current_feature)
		layout.addWidget(self.add_btn)

		self.combo_box.setCurrentIndex(0)
		self.update_arguments()

	def add_current_feature(self):

		filtername = self.combo_box.currentText()
		self.parent_window.list_widget.addItems([filtername])

		filter_instructions = [filtername.split('_')[0]]
		for a in self.arguments_le:
			arg = a.text()
			arg_num = arg
			if (arg != '') and arg_num.replace('.', '').replace(',', '').isnumeric():
				num = float(arg)
				if num.is_integer():
					num = int(num)
				filter_instructions.append(num)
			elif arg != '':
				filter_instructions.append(arg)

		print(f'You added filter {filter_instructions}.')

		self.parent_window.items.append(filter_instructions)
		self.close()

	def update_arguments(self):

		selected_filter = self.combo_box.currentText()
		arguments = self.default_params[selected_filter]
		if arguments is not None:
			args = list(arguments.keys())
			for i in range(len(args)):
				self.arguments_labels[i].setEnabled(True)
				self.arguments_le[i].setEnabled(True)

				self.arguments_labels[i].setText(args[i])
				self.arguments_le[i].setText(str(arguments[args[i]]))

			if len(args) < 2:
				for i in range(len(args), 2):
					self.arguments_labels[i].setEnabled(False)
					self.arguments_labels[i].setText('')
					self.arguments_le[i].setEnabled(False)
		else:
			for i in range(2):
				self.arguments_labels[i].setEnabled(False)
				self.arguments_le[i].setEnabled(False)
				self.arguments_labels[i].setText('')


class OperationChoice(QWidget):
	"""
	Mini window to select an operation from numpy to apply on the ROI.

	"""

	def __init__(self, parent_window):
		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Add feature")
		# Create the QComboBox and add some items
		self.combo_box = QComboBox(self)
		center_window(self)

		self.combo_box.addItems(["mean", "median", "average", "std", "var",
								 "nanmedian", "nanmean", "nanstd", "nanvar"])

		self.add_btn = QPushButton("Add")
		self.add_btn.clicked.connect(self.add_current_feature)

		# Create the layout
		layout = QVBoxLayout(self)
		layout.addWidget(self.combo_box)
		layout.addWidget(self.add_btn)

	def add_current_feature(self):
		filtername = self.combo_box.currentText()
		self.parent_window.list_widget.addItems([filtername])
		self.close()


class GeometryChoice(QWidget):

	def __init__(self, parent_window):

		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Set distances")
		center_window(self)

		# Create the QComboBox and add some items

		self.dist_label = QLabel('Distance [px]: ')
		self.dist_le = QLineEdit('10')

		self.dist_outer_label = QLabel('Max distance [px]')
		self.dist_outer_le = QLineEdit('100')
		self.outer_to_hide = [self.dist_outer_le, self.dist_outer_label]

		self.outer_btn = QCheckBox('outer distance')
		self.outer_btn.clicked.connect(self.activate_outer_value)

		self.add_btn = QPushButton("Add")
		self.add_btn.clicked.connect(self.add_current_feature)

		# Create the layout
		layout = QVBoxLayout(self)
		dist_layout = QHBoxLayout()
		dist_layout.addWidget(self.dist_label, 30)
		dist_layout.addWidget(self.dist_le, 70)

		self.dist_outer_layout = QHBoxLayout()
		self.dist_outer_layout.addWidget(self.dist_outer_label, 30)
		self.dist_outer_layout.addWidget(self.dist_outer_le, 70)

		layout.addLayout(dist_layout)
		layout.addLayout(self.dist_outer_layout)
		layout.addWidget(self.outer_btn)
		layout.addWidget(self.add_btn)

		for el in self.outer_to_hide:
			el.hide()

	def activate_outer_value(self):
		if self.outer_btn.isChecked():
			self.dist_label.setText('Min distance [px]: ')
			for el in self.outer_to_hide:
				el.show()
		else:
			self.dist_label.setText('Distance [px]: ')
			for el in self.outer_to_hide:
				el.hide()

	def add_current_feature(self):

		value = self.dist_le.text()
		if self.outer_btn.isChecked():
			value2 = self.dist_outer_le.text()
			values = [value + '-' + value2]
		else:
			values = [value]
		self.parent_window.list_widget.addItems(values)
		self.close()


class DistanceChoice(QWidget):

	def __init__(self, parent_window):
		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Set distances")
		center_window(self)

		# Create the QComboBox and add some items

		self.dist_label = QLabel('Distance [px]: ')
		self.dist_le = QLineEdit('10')

		self.add_btn = QPushButton("Add")
		self.add_btn.clicked.connect(self.add_current_feature)

		# Create the layout
		layout = QVBoxLayout(self)
		dist_layout = QHBoxLayout()
		dist_layout.addWidget(self.dist_label, 30)
		dist_layout.addWidget(self.dist_le, 70)

		layout.addLayout(dist_layout)
		layout.addWidget(self.add_btn)

	def add_current_feature(self):
		value = self.dist_le.text()
		values = [value]
		self.parent_window.list_widget.addItems(values)
		self.close()


class ListWidget(QWidget):

	"""
	A customizable widget for displaying and managing a list of items, with the 
	ability to add and remove items interactively.

	This widget is built around a `QListWidget` and allows for initialization with 
	a set of features. It also provides options to retrieve the items, add new items 
	using a custom widget, and remove selected items. The items can be parsed and 
	returned as a list, with support for various data types and formatted input (e.g., 
	ranges specified with a dash).

	Parameters
	----------
	choiceWidget : QWidget
		A custom widget that is used to add new items to the list.
	initial_features : list
		A list of initial items to populate the list widget.
	dtype : type, optional
		The data type to cast the list items to. Default is `str`.

	Attributes
	----------
	initial_features : list
		The initial set of features or items displayed in the list.
	choiceWidget : QWidget
		The widget used to prompt the user to add new items.
	dtype : type
		The data type to convert items into when retrieved from the list.
	items : list
		A list to store the current items in the list widget.
	list_widget : QListWidget
		The core Qt widget that displays the list of items.

	Methods
	-------
	addItem()
		Opens a new window to add an item to the list using the custom `choiceWidget`.
	getItems()
		Retrieves the items from the list widget, parsing ranges (e.g., 'min-max') 
		into two values, and converts them to the specified `dtype`.
	removeSel()
		Removes the currently selected item(s) from the list widget and updates the 
		internal `items` list accordingly.
	"""

	def __init__(self, choiceWidget, initial_features, dtype=str, *args, **kwargs):

		super().__init__()
		self.initial_features = initial_features
		self.choiceWidget = choiceWidget
		self.dtype = dtype
		self.items = []

		self.setFixedHeight(80)

		# Initialize list widget
		self.list_widget = QListWidget()
		self.list_widget.addItems(initial_features)

		# Set up layout
		main_layout = QVBoxLayout()
		main_layout.addWidget(self.list_widget)
		self.setLayout(main_layout)
		center_window(self)

	def addItem(self):

		"""
		Opens the custom choiceWidget to add a new item to the list.
		"""

		self.addItemWindow = self.choiceWidget(self)
		self.addItemWindow.show()

	def getItems(self):

		"""
		Retrieves and returns the items from the list widget.

		This method parses any items that contain a range (formatted as 'min-max') 
		into a list of two values, and casts all items to the specified `dtype`.

		Returns
		-------
		list
			A list of the items in the list widget, with ranges split into two values.
		"""

		items = []
		for x in range(self.list_widget.count()):
			if len(self.list_widget.item(x).text().split('-')) == 2:
				if self.list_widget.item(x).text()[0] == '-':
					items.append(self.dtype(self.list_widget.item(x).text()))
				else:
					minn, maxx = self.list_widget.item(x).text().split('-')
					to_add = [self.dtype(minn), self.dtype(maxx)]
					items.append(to_add)
			else:
				items.append(self.dtype(self.list_widget.item(x).text()))
		return items


	def removeSel(self):
		
		"""
		Removes the selected item(s) from the list widget.

		If there are any selected items, they are removed both from the visual list 
		and the internal `items` list that tracks the current state of the widget.
		"""

		listItems = self.list_widget.selectedItems()
		if not listItems: return
		for item in listItems:
			idx = self.list_widget.row(item)
			self.list_widget.takeItem(idx)
			if self.items:
				del self.items[idx]


class FigureCanvas(QWidget):
	"""
	Generic figure canvas.
	"""

	def __init__(self, fig, title="", interactive=True):
		super().__init__()
		self.fig = fig
		self.setWindowTitle(title)
		self.canvas = FigureCanvasQTAgg(self.fig)
		self.canvas.setStyleSheet("background-color: transparent;")
		if interactive:
			self.toolbar = NavigationToolbar2QT(self.canvas)
		self.layout = QVBoxLayout(self)
		self.layout.addWidget(self.canvas)
		if interactive:
			self.layout.addWidget(self.toolbar)

		center_window(self)
		self.setAttribute(Qt.WA_DeleteOnClose)

	def draw(self):
		self.canvas.draw()

	def closeEvent(self, event):
		""" Delete figure on closing window. """
		# self.canvas.ax.cla() # ****
		self.fig.clf()  # ****
		plt.close(self.fig)
		super(FigureCanvas, self).closeEvent(event)

class ThresholdLineEdit(QLineEdit):
	
	"""
	A custom QLineEdit widget to manage and validate threshold values.

	This class extends QLineEdit to input and manage threshold values (either float or int), 
	with optional validation and interaction with connected QPushButtons. The widget can 
	validate the input and enable/disable buttons based on whether a valid threshold is set.

	Parameters
	----------
	init_value : float or int, optional
		The initial threshold value to display in the input field (default is 2.0).
	connected_buttons : QPushButton or list of QPushButton, optional
		QPushButton(s) that should be enabled/disabled based on the validity of the threshold 
		value (default is None).
	placeholder : str, optional
		Placeholder text to show when no value is entered in the input field 
		(default is 'px > thresh are masked').
	value_type : str, optional
		Specifies the type of threshold value, either 'float' or 'int' (default is 'float').

	Methods
	-------
	enable_btn():
		Enables or disables connected QPushButtons based on the validity of the threshold value.
	set_threshold(value):
		Sets the input field to the given threshold value.
	get_threshold(show_warning=True):
		Retrieves the current threshold value from the input field, returning it as a float or int.
		If invalid, optionally displays a warning dialog.

	Example
	-------
	>>> threshold_input = ThresholdLineEdit(init_value=5, value_type='int')
	>>> print(threshold_input.get_threshold())
	5
	"""
	

	def __init__(self, init_value=2.0, connected_buttons=None, placeholder='px > thresh are masked',value_type='float',*args):
		super().__init__(*args)

		self.init_value = init_value
		self.value_type = value_type 
		self.connected_buttons = connected_buttons
		self.setPlaceholderText(placeholder)

		if self.value_type=="float":
			self.setValidator(QDoubleValidator())
		else:
			self.init_value = int(self.init_value)
			self.setValidator(QIntValidator())

		if self.connected_buttons is not None:
			self.textChanged.connect(self.enable_btn)
		self.set_threshold(self.init_value)

	def enable_btn(self):

		"""
		Enable or disable connected QPushButtons based on the threshold value.

		If the current threshold value is valid, the connected buttons will be enabled.
		If the value is invalid or empty, the buttons will be disabled.
		"""

		thresh = self.get_threshold(show_warning=False)
		if isinstance(self.connected_buttons, QPushButton):
			cbs = [self.connected_buttons]
		else:
			cbs = self.connected_buttons

		if thresh is None:
			for c in cbs:
				c.setEnabled(False)
		else:
			for c in cbs:
				c.setEnabled(True)

	def set_threshold(self, value):

		"""
		Set the input field to the specified threshold value.

		Parameters
		----------
		value : float or int
			The value to set in the input field.
		"""

		try:
			self.setText(str(value).replace('.',','))
		except:
			print('Please provide a valid threshold value...')

	def get_threshold(self, show_warning=True):
		
		"""
		Retrieve the current threshold value from the input field.

		Converts the value to a float or int based on the `value_type` attribute. If the value
		is invalid and `show_warning` is True, a warning dialog is shown.

		Parameters
		----------
		show_warning : bool, optional
			If True, show a warning dialog if the value is invalid (default is True).

		Returns
		-------
		float or int or None
			The threshold value as a float or int, or None if the value is invalid.
		"""

		try:
			if self.value_type=='float':
				thresh = float(self.text().replace(',','.'))
			else:
				thresh = int(self.text().replace(',','.'))
		except ValueError:
			if show_warning:
				msgBox = QMessageBox()
				msgBox.setWindowTitle('warning')
				msgBox.setIcon(QMessageBox.Critical)
				msgBox.setText("Please set a valid threshold value.")
				msgBox.setWindowTitle("")
				msgBox.setStandardButtons(QMessageBox.Ok)
				returnValue = msgBox.exec()
			thresh = None

		return thresh


def color_from_status(status, recently_modified=False):
	if not recently_modified:
		if status == 0:
			return 'tab:blue'
		elif status == 1:
			return 'tab:red'
		elif status == 2:
			return 'yellow'
		else:
			return 'k'
	else:
		if status == 0:
			return 'tab:cyan'
		elif status == 1:
			return 'tab:orange'
		elif status == 2:
			return 'tab:olive'
		else:
			return 'k'

def color_from_state(state):

	"""
	Generate a color map based on unique values in the provided state array.

	This function creates a mapping between the unique values found in the `state` array 
	and colors from the `tab10` colormap in Matplotlib. A special condition is applied 
	to the value `99`, which is assigned the color black ('k').

	Parameters
	----------
	state : array-like
		An array or list of state values to be used for generating the color map.

	Returns
	-------
	dict
		A dictionary where the keys are the unique state values from the `state` array, 
		and the values are the corresponding colors from Matplotlib's `tab10` colormap.
		The value `99` is mapped to the color black ('k').
	
	Notes
	-----
	- Matplotlib's `tab10` colormap is used for values other than `99`.
	"""

	unique_values = np.unique(state)
	color_map={}
	for value in unique_values:
		color_map[value] = plt.cm.tab10(value)
		if value == 99:
			color_map[value] = 'k'

	return color_map


def color_from_class(cclass, recently_modified=False):
	if not recently_modified:
		if cclass == 0:
			return 'tab:red'
		elif cclass == 1:
			return 'tab:blue'
		elif cclass == 2:
			return 'yellow'
		else:
			return 'k'
	else:
		if cclass == 0:
			return 'tab:orange'
		elif cclass == 1:
			return 'tab:cyan'
		elif cclass == 2:
			return 'tab:olive'
		else:
			return 'k'


class ChannelChoice(QWidget):

	def __init__(self, parent_window):
		super().__init__()
		self.parent_window = parent_window
		#self.channel_names = channel_names
		self.setWindowTitle("Choose target channel")
		# Create the QComboBox and add some items
		self.combo_box = QComboBox(self)
		center_window(self)

		channels = parent_window.channel_names

		self.combo_box.addItems(channels)

		self.add_btn = QPushButton("Add")
		self.add_btn.clicked.connect(self.add_current_channel)

		# Create the layout
		layout = QVBoxLayout(self)
		layout.addWidget(self.combo_box)
		layout.addWidget(self.add_btn)

	def add_current_channel(self):
		filtername = self.combo_box.currentText()
		self.parent_window.list_widget.addItems([filtername])
		self.close()

def help_generic(tree):

	"""
	Interactively traverse a decision tree to provide user guidance based on a nested dictionary structure.

	This function takes a nested dictionary representing a decision tree and guides the user through
	it step-by-step by displaying messages for user input using the `generic_msg()` function. 
	At each step, the user selects a key that corresponds to a further step in the tree, until a 
	final suggestion (leaf node) is reached.

	Parameters
	----------
	tree : dict
		A dictionary where keys represent options and values represent either further steps (as dictionaries)
		or a final suggestion (leaf nodes).

	Returns
	-------
	any
		The final suggestion or outcome after traversing the decision tree.

	Example
	-------
	>>> decision_tree = {
	...     'Start': {
	...         'Option 1': {
	...             'Sub-option 1': 'Final suggestion 1',
	...             'Sub-option 2': 'Final suggestion 2'
	...         },
	...         'Option 2': 'Final suggestion 3'
	...     }
	... }
	>>> result = help_generic(decision_tree)
	# The function prompts the user to choose between "Option 1" or "Option 2", 
	# and then proceeds through the tree based on the user's choices.
	"""

	output = generic_msg(list(tree.keys())[0])
	while output is not None:
		tree = tree[list(tree.keys())[0]][output]
		if isinstance(tree,dict):
			output = generic_msg(list(tree.keys())[0])
		else:
			# return the final suggestion
			output = None
	return tree

def generic_msg(text):
	
	"""
	Display a message box with a question and capture the user's response.

	This function creates a message box with a `Yes`, `No`, and `Cancel` option, 
	displaying the provided `text` as the question. It returns the user's selection as a string.

	Parameters
	----------
	text : str
		The message or question to display in the message box.

	Returns
	-------
	str or None
		The user's response: "yes" if Yes is selected, "no" if No is selected, 
		and `None` if Cancel is selected or the dialog is closed.

	Example
	-------
	>>> response = generic_msg("Would you like to continue?")
	>>> if response == "yes":
	...     print("User chose Yes")
	... elif response == "no":
	...     print("User chose No")
	... else:
	...     print("User cancelled the action")
	
	Notes
	-----
	- The message box displays a window with three options: Yes, No, and Cancel.
	"""

	msgBox = QMessageBox()
	msgBox.setIcon(QMessageBox.Question)
	msgBox.setText(text)
	msgBox.setWindowTitle("Question")
	msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
	returnValue = msgBox.exec()
	if returnValue == QMessageBox.Yes:
		return "yes"
	elif returnValue == QMessageBox.No:
		return "no"
	else:
		return None


from .styles import Styles
from .btrack_options import ConfigTracking
from .json_readers import ConfigEditor
from .tableUI import TableUI
from .measurement_options import ConfigMeasurements
from .neighborhood_options import ConfigNeighborhoods
from .classifier_widget import ClassifierWidget
from .survival_ui import ConfigSurvival
from .plot_signals_ui import ConfigSignalPlot
from .signal_annotator_options import ConfigSignalAnnotator
from .signal_annotator import SignalAnnotator
from .signal_annotator2 import SignalAnnotator2
from .retrain_signal_model_options import ConfigSignalModelTraining
from .retrain_segmentation_model_options import ConfigSegmentationModelTraining
from .thresholds_gui import ThresholdConfigWizard
from .seg_model_loader import SegmentationModelLoader
from .process_block import ProcessPanel, NeighPanel, PreprocessingPanel
from .analyze_block import AnalysisPanel
from .control_panel import ControlPanel
from .configure_new_exp import ConfigNewExperiment




from PyQt5.QtWidgets import QMessageBox,QGridLayout, QButtonGroup, \
	QCheckBox, QLineEdit, QVBoxLayout, QWidget, QLabel, QHBoxLayout, QPushButton, \
	QRadioButton
from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QDoubleValidator

from celldetective.gui.gui_utils import center_window, FigureCanvas, ExportPlotBtn
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
import numpy as np
import json
import os
import matplotlib.pyplot as plt
plt.rcParams['svg.fonttype'] = 'none'
from glob import glob
from matplotlib.cm import tab10
from celldetective.gui import Styles
import matplotlib.cm as mcm


class GenericSignalPlotWidget(QWidget, Styles):

	def __init__(self, df = None, df_pos_info = None, df_well_info = None, feature_selected = None, parent_window=None, title='plot', *args, **kwargs):
		
		super().__init__()
		center_window(self)

		self.parent_window = parent_window
		self.setWindowTitle(title)
		self.setWindowIcon(self.celldetective_icon)

		self.show_ci = False
		self.legend_visible = True
		self.show_cell_lines = False
		self.alpha_setting = 0.5
		self.scaling_factor = 1
		self.target_class = [0]
		self.float_validator = QDoubleValidator()

		cmap_lbl = self.parent_window.cbs[-1].currentText()
		self.cmap = getattr(mcm, cmap_lbl)

		self.feature_selected = feature_selected
		self.df = df
		self.df_pos_info = df_pos_info
		self.df_well_info = df_well_info

		self.layout = QVBoxLayout()
		self.layout.setSpacing(3)
		self.populate_widget()

		self.ci_btn.click()
		#self.legend_btn.click()

		self.fig.tight_layout()
		self.setLayout(self.layout)
		self.setAttribute(Qt.WA_DeleteOnClose)

	def populate_widget(self):

		self.plot_options = [QRadioButton() for i in range(3)]
		self.radio_labels = ['well', 'position', 'both']
		radio_hbox = QHBoxLayout()
		radio_hbox.setContentsMargins(15,15,15,0)
		
		self.group_lbl = QLabel('grouping: ')
		radio_hbox.addWidget(self.group_lbl, 25)
		radio_subhbox = QHBoxLayout()
		radio_hbox.addLayout(radio_subhbox, 75)

		self.plot_btn_group = QButtonGroup()
		for i in range(3):
			self.plot_options[i].setText(self.radio_labels[i])
			self.plot_btn_group.addButton(self.plot_options[i])
			radio_subhbox.addWidget(self.plot_options[i], 33, alignment=Qt.AlignCenter)

		if self.parent_window.position_indices is not None:
			if len(self.parent_window.well_indices)>1 and len(self.parent_window.position_indices)==1:
				self.plot_btn_group.buttons()[0].click()
				for i in [1,2]:
					self.plot_options[i].setEnabled(False)
			elif len(self.parent_window.well_indices)>1:
				self.plot_btn_group.buttons()[0].click()
			elif len(self.parent_window.well_indices)==1 and len(self.parent_window.position_indices)==1:
				self.plot_btn_group.buttons()[1].click()
				for i in [0,2]:
					self.plot_options[i].setEnabled(False)
		else:
			if len(self.parent_window.well_indices)>1:
				self.plot_btn_group.buttons()[0].click()
			elif len(self.parent_window.well_indices)==1:
				self.plot_btn_group.buttons()[2].click()

		self.layout.addLayout(radio_hbox)


		plot_buttons_hbox = QHBoxLayout()
		plot_buttons_hbox.setContentsMargins(10,10,5,0)
		plot_buttons_hbox.addWidget(QLabel(''),80, alignment=Qt.AlignLeft)

		self.legend_btn = QPushButton('')
		self.legend_btn.setIcon(icon(MDI6.text_box,color=self.help_color))
		self.legend_btn.setStyleSheet(self.button_select_all)
		self.legend_btn.setToolTip('Show or hide the legend')
		self.legend_btn.setIconSize(QSize(20, 20))
		plot_buttons_hbox.addWidget(self.legend_btn, 5,alignment=Qt.AlignRight)

		self.log_btn = QPushButton('')
		self.log_btn.setIcon(icon(MDI6.math_log,color="black"))
		self.log_btn.setStyleSheet(self.button_select_all)
		self.log_btn.setIconSize(QSize(20, 20))
		self.log_btn.setToolTip('Enable or disable log scale')
		plot_buttons_hbox.addWidget(self.log_btn, 5, alignment=Qt.AlignRight)

		self.ci_btn = QPushButton('')
		self.ci_btn.setIcon(icon(MDI6.arrow_expand_horizontal,color="black"))
		self.ci_btn.setStyleSheet(self.button_select_all)
		self.ci_btn.setIconSize(QSize(20, 20))
		self.ci_btn.setToolTip('Show or hide confidence intervals.')
		plot_buttons_hbox.addWidget(self.ci_btn, 5, alignment=Qt.AlignRight)

		self.cell_lines_btn = QPushButton('')
		self.cell_lines_btn.setIcon(icon(MDI6.view_headline,color="black"))
		self.cell_lines_btn.setStyleSheet(self.button_select_all)
		self.cell_lines_btn.setToolTip('Show or hide individual cell signals.')
		self.cell_lines_btn.setIconSize(QSize(20, 20))
		plot_buttons_hbox.addWidget(self.cell_lines_btn, 5, alignment=Qt.AlignRight)

		self.fig, self.ax = plt.subplots(1,1,figsize=(4,3))
		self.plot_widget = FigureCanvas(self.fig, title="")
		self.plot_widget.setContentsMargins(0,0,0,0)
		self.initialize_axis()
		plt.tight_layout()

		self.export_btn = ExportPlotBtn(self.fig, export_dir=self.parent_window.exp_dir)
		plot_buttons_hbox.addWidget(self.export_btn, 5, alignment=Qt.AlignRight)
		self.layout.addLayout(plot_buttons_hbox)

		self.ax.set_prop_cycle('color',[self.cmap(i) for i in np.linspace(0, 1, len(self.parent_window.well_indices))])

		self.fig.set_facecolor('none')  # or 'None'
		self.fig.canvas.setStyleSheet("background-color: transparent;")
		self.plot_widget.canvas.draw()

		self.layout.addWidget(self.plot_widget)

		self.plot_btn_group.buttonClicked[int].connect(self.plot_signals)
		self.legend_btn.clicked.connect(self.show_hide_legend)
		self.log_btn.clicked.connect(self.switch_to_log)
		self.ci_btn.clicked.connect(self.switch_ci)
		self.cell_lines_btn.clicked.connect(self.switch_cell_lines)

		self.class_selection_widget = QWidget()

		self.class_selection_lbl = QLabel('class of interest:')
		class_hbox = QHBoxLayout()
		self.class_selection_widget.setLayout(class_hbox)
		class_hbox.addWidget(self.class_selection_lbl, 25, alignment=Qt.AlignLeft)

		class_subhbox = QHBoxLayout()
		class_hbox.addLayout(class_subhbox, 75)

		self.all_btn = QRadioButton('*')
		self.event_btn = QRadioButton('event')
		self.event_btn.setChecked(True)
		self.no_event_btn = QRadioButton('no event')
		self.class_btn_group = QButtonGroup()
		for btn in [self.all_btn, self.event_btn, self.no_event_btn]:
			self.class_btn_group.addButton(btn)

		self.class_btn_group.buttonClicked[int].connect(self.set_class_to_plot)

		class_subhbox.addWidget(self.all_btn, 33, alignment=Qt.AlignCenter)
		class_subhbox.addWidget(self.event_btn, 33, alignment=Qt.AlignCenter)
		class_subhbox.addWidget(self.no_event_btn, 33, alignment=Qt.AlignCenter)

		self.layout.addWidget(self.class_selection_widget) #Layout(class_hbox)

		# Rescale 
		self.rescale_widget = QWidget()

		scale_hbox = QHBoxLayout()
		self.rescale_widget.setLayout(scale_hbox)

		scale_hbox.addWidget(QLabel('scaling factor: '), 25)
		self.scaling_factor_le = QLineEdit('1')
		self.scaling_factor_le.setValidator(self.float_validator)
		scale_hbox.addWidget(self.scaling_factor_le, 65)

		self.rescale_btn = QPushButton('rescale')
		self.rescale_btn.setStyleSheet(self.button_style_sheet_2)
		self.rescale_btn.clicked.connect(self.rescale_y_axis)
		scale_hbox.addWidget(self.rescale_btn, 10)
		#self.layout.addLayout(scale_hbox)
		self.layout.addWidget(self.rescale_widget)


		# Rescale 
		self.cell_lines_alpha_wdg = QWidget()
		alpha_hbox = QHBoxLayout()
		self.cell_lines_alpha_wdg.setLayout(alpha_hbox)

		alpha_hbox.addWidget(QLabel('single-cell\nsignal alpha: '), 25)
		self.alpha_le = QLineEdit('0,8')
		self.alpha_le.setValidator(self.float_validator)
		alpha_hbox.addWidget(self.alpha_le, 65)

		self.submit_alpha_btn = QPushButton('submit')
		self.submit_alpha_btn.setStyleSheet(self.button_style_sheet_2)
		self.submit_alpha_btn.clicked.connect(self.submit_alpha)
		alpha_hbox.addWidget(self.submit_alpha_btn, 10)
		self.layout.addWidget(self.cell_lines_alpha_wdg)

		self.select_option = [QRadioButton() for i in range(2)]
		self.select_label = ['by name', 'spatially']

		select_hbox = QHBoxLayout()
		select_hbox.addWidget(QLabel('select position: '), 25)

		select_subhbox = QHBoxLayout()
		select_hbox.addLayout(select_subhbox, 75)

		self.select_btn_group = QButtonGroup()
		for i in range(2):
			self.select_option[i].setText(self.select_label[i])
			self.select_btn_group.addButton(self.select_option[i])
			select_subhbox.addWidget(self.select_option[i],33, alignment=Qt.AlignCenter)
		self.select_option[0].setChecked(True)
		self.select_btn_group.buttonClicked[int].connect(self.switch_selection_mode)
		self.layout.addLayout(select_hbox)

		self.look_for_metadata()
		if self.metadata_found:
			self.fig_scatter, self.ax_scatter = plt.subplots(1,1,figsize=(4,2)) #,figsize=(4,3)
			self.position_scatter = FigureCanvas(self.fig_scatter)
			self.load_coordinates()
			self.plot_spatial_location()
			self.ax_scatter.spines['top'].set_visible(False)
			self.ax_scatter.spines['right'].set_visible(False)
			self.ax_scatter.set_aspect('equal')
			self.ax_scatter.set_xticks([])
			self.ax_scatter.set_yticks([])
			plt.tight_layout()

			self.fig_scatter.set_facecolor('none')  # or 'None'
			self.fig_scatter.canvas.setStyleSheet("background-color: transparent;")
			self.layout.addWidget(self.position_scatter)

		self.generate_pos_selection_widget()
		self.select_btn_group.buttons()[0].click()

	def submit_alpha(self):

		alpha = self.alpha_le.text().replace(',','.')
		try:
			alpha = float(alpha)
		except:
			return None
		if alpha>1.0:
			alpha = 1.0
		elif alpha<0.0:
			alpha = 0.0
		self.alpha_setting = alpha
		self.plot_signals(0)

	def rescale_y_axis(self):
		new_scale = self.scaling_factor_le.text().replace(',','.')
		if new_scale=='':
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("Please set a valid scaling factor...")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None
		else:
			self.scaling_factor = float(new_scale)
			self.plot_signals(0)


	def switch_selection_mode(self, id):

		for i in range(2):
			if self.select_option[i].isChecked():
				self.selection_mode = self.select_label[i]
		if self.selection_mode=='by name':
			if len(self.metafiles)>0:
				self.position_scatter.hide()
			self.line_choice_widget.show()
		else:
			if len(self.metafiles)>0:
				self.position_scatter.show()
			self.line_choice_widget.hide()

	def set_class_to_plot(self):

		if self.all_btn.isChecked():
			self.target_class=[0,1]
		elif self.event_btn.isChecked():
			self.target_class = [0]
		else:
			self.target_class = [1]

		self.plot_signals(0)

	def generate_pos_selection_widget(self):
		
		self.well_names = self.df['well_name'].unique()
		self.pos_names = self.df_pos_info['pos_name'].unique() #pd.DataFrame(self.ks_estimators_per_position)['position_name'].unique()

		self.usable_well_labels = []
		for name in self.well_names:
			for lbl in self.parent_window.well_labels:
				if name+':' in lbl:
					self.usable_well_labels.append(lbl)

		self.line_choice_widget = QWidget()
		self.line_check_vbox = QGridLayout()
		self.line_choice_widget.setLayout(self.line_check_vbox)

		if len(self.parent_window.well_indices)>1:
			self.well_display_options = [QCheckBox(self.usable_well_labels[i]) for i in range(len(self.usable_well_labels))]
			for i in range(len(self.well_names)):
				self.line_check_vbox.addWidget(self.well_display_options[i], i, 0, 1, 1, alignment=Qt.AlignLeft)
				self.well_display_options[i].setChecked(True)
				self.well_display_options[i].setStyleSheet("font-size: 12px;")
				self.well_display_options[i].toggled.connect(self.select_lines)
		else:
			self.pos_display_options = [QCheckBox(self.pos_names[i]) for i in range(len(self.pos_names))]
			for i in range(len(self.pos_names)):
				self.line_check_vbox.addWidget(self.pos_display_options[i], i%4, i//4, 1, 1, alignment=Qt.AlignCenter)
				self.pos_display_options[i].setChecked(True)
				self.pos_display_options[i].setStyleSheet("font-size: 12px;")
				self.pos_display_options[i].toggled.connect(self.select_lines)

		self.layout.addWidget(self.line_choice_widget)
		#self.layout.addLayout(self.line_check_vbox)

	def look_for_metadata(self):
		self.metadata_found = False
		self.metafiles = glob(self.parent_window.exp_dir+os.sep.join([f'W*','*','movie','*metadata.txt'])) \
					+ glob(self.parent_window.exp_dir+os.sep.join([f'W*','*','*metadata.txt'])) \
					+ glob(self.parent_window.exp_dir+os.sep.join([f'W*','*metadata.txt'])) \
					+ glob(self.parent_window.exp_dir+'*metadata.txt')
		print(f'Found {len(self.metafiles)} metadata files...')
		if len(self.metafiles)>0:
			self.metadata_found = True

	def load_coordinates(self):

		"""
		Read metadata and try to extract position coordinates
		"""

		self.no_meta = False
		try:
			with open(self.metafiles[0], 'r') as f:
				data = json.load(f)
				positions = data['Summary']['InitialPositionList']
		except Exception as e:
			print(f'Trouble loading metadata: error {e}...')
			return None

		for k in range(len(positions)):
			pos_label = positions[k]['Label']
			try:
				coords = positions[k]['DeviceCoordinatesUm']['XYStage']
			except:
				try:
					coords = positions[k]['DeviceCoordinatesUm']['PIXYStage']
				except:
					self.no_meta = True

			if not self.no_meta:
				files = self.df_pos_info['stack_path'].values
				pos_loc = [pos_label in f for f in files]
				self.df_pos_info.loc[pos_loc, 'x'] = coords[0]
				self.df_pos_info.loc[pos_loc, 'y'] = coords[1]
				self.df_pos_info.loc[pos_loc, 'metadata_tag'] = pos_label


	def plot_spatial_location(self):

		try:
			self.sc = self.ax_scatter.scatter(self.df_pos_info["x"].values, self.df_pos_info["y"].values, picker=True, pickradius=1, color=self.select_color(self.df_pos_info["select"].values))
			self.scat_labels = self.df_pos_info['metadata_tag'].values
			self.ax_scatter.invert_xaxis()

			self.annot = self.ax_scatter.annotate("", xy=(0,0), xytext=(10,10),textcoords="offset points",
								bbox=dict(boxstyle="round", fc="w"),
								arrowprops=dict(arrowstyle="->"))
			self.annot.set_visible(False)
			
			xmin,xmax = self.ax_scatter.get_xlim()
			ymin,ymax = self.ax_scatter.get_ylim()
			xdatarange = xmax - xmin
			ydatarange = ymax - ymin
			
			self.ax_scatter.set_xlim(xmin-0.1*xdatarange, xmax + 0.1*xdatarange)
			self.ax_scatter.set_ylim(ymin-0.1*ydatarange, ymax + 0.1*ydatarange)
			
			#xmin,xmax = self.ax_scatter.get_xlim()
			#ymin,ymax = self.ax_scatter.get_ylim()
			#desired_a = 4
			#new_x_max = xmin + desired_a * (ymax - ymin)
			#self.ax_scatter.set_xlim(xmin - (new_x_max - xmin)/2.0, new_x_max - (new_x_max - xmin)/2.0)
			
			self.fig_scatter.tight_layout()
			self.fig_scatter.canvas.mpl_connect("motion_notify_event", self.hover)
			self.fig_scatter.canvas.mpl_connect("pick_event", self.unselect_position)
		except Exception as e:
			pass

	def update_annot(self, ind):
		
		pos = self.sc.get_offsets()[ind["ind"][0]]
		self.annot.xy = pos
		text = self.scat_labels[ind["ind"][0]]
		self.annot.set_text(text)
		self.annot.get_bbox_patch().set_facecolor('k')
		self.annot.get_bbox_patch().set_alpha(0.4)


	def hover(self, event):
		vis = self.annot.get_visible()
		if event.inaxes == self.ax_scatter:
			cont, ind = self.sc.contains(event)
			if cont:
				self.update_annot(ind)
				self.annot.set_visible(True)
				self.fig_scatter.canvas.draw_idle()
			else:
				if vis:
					self.annot.set_visible(False)
					self.fig_scatter.canvas.draw_idle()

	def unselect_position(self, event):
		
		ind = event.ind # index of selected position
		well_idx = self.df_pos_info.iloc[ind]['well_index'].values[0]
		selectedPos = self.df_pos_info.iloc[ind]['pos_path'].values[0]
		currentSelState = self.df_pos_info.iloc[ind]['select'].values[0]
		if self.plot_options[0].isChecked() or self.plot_options[2].isChecked():
			self.df_pos_info.loc[self.df_pos_info['well_index']==well_idx,'select'] = not currentSelState
			self.df_well_info.loc[self.df_well_info['well_index']==well_idx, 'select'] = not currentSelState
			if len(self.parent_window.well_indices)>1:
				self.well_display_options[well_idx].setChecked(not currentSelState)
			else:
				for p in self.pos_display_options:
					p.setChecked(not currentSelState)
		else:
			self.df_pos_info.loc[self.df_pos_info['pos_path']==selectedPos,'select'] = not currentSelState
			if len(self.parent_window.well_indices)<=1:
				self.pos_display_options[ind[0]].setChecked(not currentSelState)

		self.sc.set_color(self.select_color(self.df_pos_info["select"].values))
		self.position_scatter.canvas.draw_idle()
		self.plot_signals(0)

	def select_color(self, selection):
		colors = [tab10(0) if s else tab10(0.1) for s in selection]
		return colors

	def initialize_axis(self):

		previous_ymin, previous_ymax = self.ax.get_ylim()
		previous_legend = self.legend_visible
		is_log = self.ax.get_yscale()

		self.ax.clear()
		self.ax.plot([],[])

		# Labels
		self.ax.set_xlabel('time [min]')
		self.ax.set_ylabel(self.feature_selected)

		# Spines
		self.ax.spines['top'].set_visible(False)
		self.ax.spines['right'].set_visible(False)

		# Lims
		safe_df = self.df.dropna(subset=self.feature_selected)
		values = safe_df[self.feature_selected].values
		if len(values)>0:
			self.ax.set_ylim(np.percentile(values, 1)*self.scaling_factor, np.percentile(values, 99)*self.scaling_factor)
		self.ax.set_xlim(-(self.df['FRAME'].max()+2)*self.parent_window.FrameToMin,(self.df['FRAME'].max()+2)*self.parent_window.FrameToMin)
		
		if is_log=='log':
			self.ax.set_yscale('log')
		if previous_legend:
			leg = self.ax.get_legend()
			if leg is not None:
				leg.set_visible(True)

	def show_hide_legend(self):
		
		if self.legend_visible:
			leg = self.ax.get_legend()
			leg.set_visible(False)
			self.legend_visible = False
			self.legend_btn.setIcon(icon(MDI6.text_box,color="black"))
		else:
			leg = self.ax.get_legend()
			leg.set_visible(True)
			self.legend_visible = True
			self.legend_btn.setIcon(icon(MDI6.text_box,color=self.help_color))

		self.plot_widget.canvas.draw_idle()

	def switch_to_log(self):

		"""
		Switch threshold histogram to log scale. Auto adjust.
		"""

		if self.ax.get_yscale()=='linear':
			self.ax.set_yscale('log')
			self.log_btn.setIcon(icon(MDI6.math_log,color=self.help_color))
			#self.ax.set_ylim(0.01,1.05)
		else:
			self.ax.set_yscale('linear')
			self.log_btn.setIcon(icon(MDI6.math_log,color="black"))
			#self.ax.set_ylim(0.01,1.05)

		#self.ax.autoscale()
		self.plot_widget.canvas.draw_idle()

	def plot_signals(self, id):

		for i in range(3):
			if self.plot_options[i].isChecked():
				self.plot_mode = self.radio_labels[i]

		if self.target_class==[0,1]:
			mean_signal_type = 'mean_all'
			std_signal = 'std_all'
			matrix = 'matrix_all'
		elif self.target_class==[0]:
			mean_signal_type = 'mean_event'
			std_signal = 'std_event'
			matrix = 'matrix_event'
		else:
			mean_signal_type = 'mean_no_event'
			std_signal = 'std_no_event'
			matrix = 'matrix_no_event'


		colors = np.array([self.cmap(i / len(self.df_pos_info)) for i in range(len(self.df_pos_info))])
		well_color = [self.cmap(i / len(self.df_well_info)) for i in range(len(self.df_well_info))]

		if self.plot_mode=='position':
			self.initialize_axis()
			lines = self.df_pos_info.loc[self.df_pos_info['select'],'signal'].values
			pos_labels = self.df_pos_info.loc[self.df_pos_info['select'],'pos_name'].values
			pos_indices = self.df_pos_info.loc[self.df_pos_info['select'],'pos_index'].values
			well_index = self.df_pos_info.loc[self.df_pos_info['select'],'well_index'].values
			for i in range(len(lines)):
				if len(self.parent_window.well_indices)<=1:
					self.plot_line(lines[i], colors[pos_indices[i]], pos_labels[i], mean_signal_type, std_signal=std_signal, ci_option=self.show_ci, cell_lines_option=self.show_cell_lines, matrix=matrix)
				else:
					self.plot_line(lines[i], well_color[well_index[i]], pos_labels[i], mean_signal_type, std_signal=std_signal, ci_option=self.show_ci, cell_lines_option=self.show_cell_lines, matrix=matrix)
			if self.legend_visible:
				self.ax.legend(ncols=3,fontsize='x-small')

		elif self.plot_mode=='well':
			self.initialize_axis()
			lines = self.df_well_info.loc[self.df_well_info['select'],'signal'].values	
			well_index = self.df_well_info.loc[self.df_well_info['select'],'well_index'].values
			well_labels = self.df_well_info.loc[self.df_well_info['select'],'well_name'].values
			for i in range(len(lines)):	
				if len(self.parent_window.well_indices)<=1:
					self.plot_line(lines[i], 'k', well_labels[i], mean_signal_type, std_signal=std_signal, ci_option=self.show_ci, cell_lines_option=self.show_cell_lines,matrix=matrix)
				else:
					self.plot_line(lines[i], well_color[well_index[i]], well_labels[i], mean_signal_type, std_signal=std_signal, ci_option=self.show_ci, cell_lines_option=self.show_cell_lines,matrix=matrix)
			if self.legend_visible:
				self.ax.legend(ncols=2, fontsize='x-small')

		elif self.plot_mode=='both':
			
			self.initialize_axis()
			lines_pos = self.df_pos_info.loc[self.df_pos_info['select'],'signal'].values
			lines_well = self.df_well_info.loc[self.df_well_info['select'],'signal'].values	

			pos_indices = self.df_pos_info.loc[self.df_pos_info['select'],'pos_index'].values
			well_index_pos = self.df_pos_info.loc[self.df_pos_info['select'],'well_index'].values
			well_index = self.df_well_info.loc[self.df_well_info['select'],'well_index'].values
			well_labels = self.df_well_info.loc[self.df_well_info['select'],'well_name'].values
			pos_labels = self.df_pos_info.loc[self.df_pos_info['select'],'pos_name'].values

			for i in range(len(lines_pos)):
				if len(self.parent_window.well_indices)<=1:
					self.plot_line(lines_pos[i], colors[pos_indices[i]], pos_labels[i], mean_signal_type, std_signal=std_signal, ci_option=self.show_ci, cell_lines_option=self.show_cell_lines,matrix=matrix)
				else:
					self.plot_line(lines_pos[i], well_color[well_index_pos[i]], None, mean_signal_type, std_signal=std_signal, ci_option=False)

			for i in range(len(lines_well)):
				if len(self.parent_window.well_indices)<=1:
					self.plot_line(lines_well[i], 'k', 'pool', mean_signal_type, std_signal=std_signal, ci_option=False)
				else:
					self.plot_line(lines_well[i], well_color[well_index[i]], well_labels[i], mean_signal_type, std_signal=std_signal, ci_option=False)
			if self.legend_visible:
				self.ax.legend(ncols=3,fontsize='x-small')

		self.plot_widget.canvas.draw()

	def plot_line(self, line, color, label, mean_signal_type, ci_option=True, cell_lines_option=False, alpha_ci=0.5, std_signal=None, matrix=None):
		
		# Plot a signal
		if line==line:
			self.ax.plot(line['timeline']*self.parent_window.FrameToMin, line[mean_signal_type]*self.scaling_factor, color=color, label=label)

			if ci_option and std_signal is not None:

				self.ax.fill_between(line['timeline']*self.parent_window.FrameToMin,
								[a-b for a,b in zip(line[mean_signal_type]*self.scaling_factor, line[std_signal]*self.scaling_factor)],
								[a+b for a,b in zip(line[mean_signal_type]*self.scaling_factor, line[std_signal]*self.scaling_factor)],
								color=color,
								alpha=alpha_ci,
								)
			if cell_lines_option and matrix is not None:
				# Show individual cell signals
				mat = line[matrix]
				for i in range(mat.shape[0]):
					self.ax.plot(line['timeline']*self.parent_window.FrameToMin, mat[i,:]*self.scaling_factor, color=color, alpha=self.alpha_setting)


	def switch_ci(self):

		# Show the confidence interval / STD

		if self.show_ci:
			self.ci_btn.setIcon(icon(MDI6.arrow_expand_horizontal,color="black"))
		else:
			self.ci_btn.setIcon(icon(MDI6.arrow_expand_horizontal,color=self.help_color))
		self.show_ci = not self.show_ci
		self.plot_signals(0)

	def switch_cell_lines(self):

		# Show individual cell signals

		if self.show_cell_lines:
			self.cell_lines_btn.setIcon(icon(MDI6.view_headline,color="black"))
		else:
			self.cell_lines_btn.setIcon(icon(MDI6.view_headline,color=self.help_color))
		self.show_cell_lines = not self.show_cell_lines
		self.plot_signals(0)

	def select_lines(self):
		
		if len(self.parent_window.well_indices)>1:
			for i in range(len(self.well_display_options)):
				self.df_well_info.loc[self.df_well_info['well_index']==i,'select'] = self.well_display_options[i].isChecked()
				self.df_pos_info.loc[self.df_pos_info['well_index']==i,'select'] = self.well_display_options[i].isChecked()
		else:
			for i in range(len(self.pos_display_options)):
				self.df_pos_info.loc[self.df_pos_info['pos_index']==i,'select'] = self.pos_display_options[i].isChecked()

		if len(self.metafiles)>0:
			self.sc.set_color(self.select_color(self.df_pos_info["select"].values))
			self.position_scatter.canvas.draw_idle()
		self.plot_signals(0)

class SurvivalPlotWidget(GenericSignalPlotWidget):

	def __init__(self, *args, **kwargs):

		super(SurvivalPlotWidget, self).__init__(*args, **kwargs)
		self.cell_lines_btn.hide()
		self.class_selection_widget.hide()
		self.rescale_widget.hide()
		self.cell_lines_alpha_wdg.hide()
		
	def switch_to_log(self):

		"""
		Switch threshold histogram to log scale. Auto adjust.
		"""

		if self.ax.get_yscale()=='linear':
			ymin,_ = self.ax.get_ylim()
			self.ax.set_ylim(max(ymin,0.01),1.05)
			self.ax.set_yscale('log')
			self.log_btn.setIcon(icon(MDI6.math_log,color=self.help_color))
		else:
			self.ax.set_yscale('linear')
			self.log_btn.setIcon(icon(MDI6.math_log,color="black"))
			#self.ax.set_ylim(0.01,1.05)

		#self.ax.autoscale()
		self.plot_widget.canvas.draw_idle()

	def initialize_axis(self):

		previous_legend = self.legend_visible
		is_log = self.ax.get_yscale()

		self.ax.clear()
		self.ax.plot([],[])

		# Labels
		self.ax.set_xlabel('time [min]')
		self.ax.set_ylabel('survival')

		# Spines
		self.ax.spines['top'].set_visible(False)
		self.ax.spines['right'].set_visible(False)

		# Lims
		self.ax.set_xlim(0*self.parent_window.FrameToMin,(self.df['FRAME'].max()+2)*self.parent_window.FrameToMin)
		if is_log=='log':
			ymin = 0.1
		else:
			ymin = 0.0
		self.ax.set_ylim(ymin,1.05)
	
		if is_log=='log':
			self.ax.set_yscale('log')
		if previous_legend:
			leg = self.ax.get_legend()
			if leg is not None:
				leg.set_visible(True)

	def plot_signals(self, id):

		for i in range(3):
			if self.plot_options[i].isChecked():
				self.plot_mode = self.radio_labels[i]

		colors = np.array([self.cmap(i / len(self.df_pos_info)) for i in range(len(self.df_pos_info))])
		well_color = [self.cmap(i / len(self.df_well_info)) for i in range(len(self.df_well_info))]

		if self.plot_mode=='position':
			self.initialize_axis()
			lines = self.df_pos_info.loc[self.df_pos_info['select'],'survival_fit'].values
			pos_labels = self.df_pos_info.loc[self.df_pos_info['select'],'pos_name'].values
			pos_indices = self.df_pos_info.loc[self.df_pos_info['select'],'pos_index'].values
			well_index = self.df_pos_info.loc[self.df_pos_info['select'],'well_index'].values
			for i in range(len(lines)):
				if len(self.parent_window.well_indices)<=1:
					self.plot_line(lines[i], colors[pos_indices[i]], pos_labels[i], ci_option=self.show_ci)
				else:
					self.plot_line(lines[i], well_color[well_index[i]], pos_labels[i], ci_option=self.show_ci)
			if self.legend_visible:
				self.ax.legend(ncols=3,fontsize='x-small')

		elif self.plot_mode=='well':
			self.initialize_axis()
			lines = self.df_well_info.loc[self.df_well_info['select'],'survival_fit'].values	
			well_index = self.df_well_info.loc[self.df_well_info['select'],'well_index'].values
			well_labels = self.df_well_info.loc[self.df_well_info['select'],'well_name'].values
			for i in range(len(lines)):	
				if len(self.parent_window.well_indices)<=1:
					self.plot_line(lines[i], 'k', well_labels[i], ci_option=self.show_ci, legend=True)
				else:
					self.plot_line(lines[i], well_color[well_index[i]], well_labels[i], ci_option=self.show_ci, legend=True)
			if self.legend_visible:
				self.ax.legend(ncols=2, fontsize='x-small')

		elif self.plot_mode=='both':
			
			self.initialize_axis()
			lines_pos = self.df_pos_info.loc[self.df_pos_info['select'],'survival_fit'].values
			lines_well = self.df_well_info.loc[self.df_well_info['select'],'survival_fit'].values	

			pos_indices = self.df_pos_info.loc[self.df_pos_info['select'],'pos_index'].values
			well_index_pos = self.df_pos_info.loc[self.df_pos_info['select'],'well_index'].values
			well_index = self.df_well_info.loc[self.df_well_info['select'],'well_index'].values
			well_labels = self.df_well_info.loc[self.df_well_info['select'],'well_name'].values
			pos_labels = self.df_pos_info.loc[self.df_pos_info['select'],'pos_name'].values

			for i in range(len(lines_pos)):
				if len(self.parent_window.well_indices)<=1:
					self.plot_line(lines_pos[i], colors[pos_indices[i]], pos_labels[i], ci_option=self.show_ci, legend=True)
				else:
					self.plot_line(lines_pos[i], well_color[well_index_pos[i]], None, ci_option=False)

			for i in range(len(lines_well)):
				if len(self.parent_window.well_indices)<=1:
					self.plot_line(lines_well[i], 'k', 'pool', ci_option=False, legend=True)
				else:
					self.plot_line(lines_well[i], well_color[well_index[i]], well_labels[i], ci_option=False, legend=True)
			if self.legend_visible:
				self.ax.legend(ncols=3,fontsize='x-small')

		self.plot_widget.canvas.draw()

	def plot_line(self, line, color, label, ci_option=True, legend=None, alpha_ci=0.5):
		
		# Plot a signal
		if line==line:
			line.plot_survival_function(ci_show=ci_option, ax=self.ax, legend=legend, color=color, label=label, xlabel='timeline [min]')


import configparser
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QScrollArea, QLabel, QHBoxLayout, QLineEdit, QPushButton
from PyQt5.QtCore import Qt
import configparser
from celldetective.gui import Styles

class ConfigEditor(QWidget, Styles):
    
    def __init__(self, parent_window):

        """
        Load and edit the experiment config.
        """

        super().__init__()

        self.parent_window = parent_window
        self.config_path = self.parent_window.exp_config

        self.setGeometry(500,200,400,700)

        self.setWindowTitle("Configuration")

        # Create the main layout
        self.layout = QVBoxLayout()

        # Create a scroll area to contain the main layout
        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll_content = QWidget()
        self.scroll_layout = QVBoxLayout(scroll_content)
        scroll_content.setLayout(self.scroll_layout)
        scroll.setWidget(scroll_content)

        # Add a label for each section of the config file
        self.labels = []
        self.edit_boxes = []
        self.save_button = None
        self.sections = {}

        # Set the main layout
        self.scroll_layout.addStretch()
        self.layout.addWidget(scroll)
        self.setLayout(self.layout)

        self.load_config()

    def load_config(self):
        file_name = self.config_path
        #self.file_edit.setText(file_name)

        config = configparser.ConfigParser()
        config.read(file_name)

        # Create a layout for each section of the config file
        for section in config.sections():
            section_layout = QVBoxLayout()
            section_label = QLabel('[{}]'.format(section))
            self.labels.append(section_label)

            # Create an editor box for each parameter in the section
            for key, value in config.items(section):
                edit_box_layout = QHBoxLayout()
                label = QLabel(key)
                edit_box = QLineEdit(value)
                edit_box_layout.addWidget(label)
                edit_box_layout.addWidget(edit_box)
                section_layout.addLayout(edit_box_layout)

                self.edit_boxes.append(edit_box)
                self.sections[key] = (section, edit_box)

            # Add the section label and editor boxes to the main layout
            self.scroll_layout.addWidget(section_label)
            self.scroll_layout.addLayout(section_layout)

        # Add a save button
        save_layout = QHBoxLayout()
        save_button = QPushButton('Save')
        save_button.setStyleSheet(self.button_style_sheet)
        save_button.clicked.connect(self.save_config)
        save_button.setShortcut("Return")
        #save_button.setIcon(QIcon_from_svg(self.parent.abs_path+f"/icons/save.svg", color='white'))

        #save_layout.addStretch()
        save_layout.addWidget(save_button, alignment=Qt.AlignTop)

        # Add the save button to the main layout
        self.layout.addLayout(save_layout)
        self.save_button = save_button

    def save_config(self):
        # Save the configuration to the file
        file_name = self.config_path
            
        config = configparser.ConfigParser()

        # Update the values in the config object



        for key, (section, edit_box) in self.sections.items():
            if not config.has_section(section):
                config.add_section(section)
            config.set(section, key, edit_box.text())

        # Write the config object to the file
        with open(file_name, 'w') as f:
            config.write(f)

        self.parent_window.load_configuration()
        self.close()


from skimage.filters import difference_of_gaussians, threshold_otsu, threshold_local, threshold_niblack, threshold_sauvola
from celldetective.utils import interpolate_nan
import scipy.ndimage as snd
import numpy as np

def gauss_filter(img, sigma, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))
	return snd.gaussian_filter(img.astype(float), sigma, *kwargs)

def median_filter(img, size, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))

	size = int(size)
	return snd.median_filter(img, size, *kwargs)

def maximum_filter(img, size, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))

	return snd.maximum_filter(img.astype(float), size, *kwargs)

def minimum_filter(img, size, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))

	return snd.minimum_filter(img.astype(float), size, *kwargs)

def percentile_filter(img, percentile, size, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))

	return snd.percentile_filter(img.astype(float), percentile, size, *kwargs)

def subtract_filter(img, value, *kwargs):
	return img.astype(float) - value

def abs_filter(img, *kwargs):
	return np.abs(img)

def ln_filter(img, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))

	img[np.where(img>0.)] = np.log(img[np.where(img>0.)])
	img[np.where(img<=0.)] = 0.

	return img

def variance_filter(img, size, interpolate=True):

	if interpolate:
		img = interpolate_nan(img.astype(float))

	size = int(size)
	img = img.astype(float)
	win_mean = snd.uniform_filter(img, (size,size), mode='wrap')
	win_sqr_mean = snd.uniform_filter(img**2, (size,size), mode='wrap')
	img = win_sqr_mean - win_mean**2

	return img

def std_filter(img, size, interpolate=True):

	if interpolate:
		img = interpolate_nan(img.astype(float))
	size = int(size)
	img = img.astype(float)
	win_mean = snd.uniform_filter(img, (size,size), mode='wrap')
	win_sqr_mean = snd.uniform_filter(img**2, (size, size), mode='wrap')
	win_sqr_mean[win_sqr_mean!=win_sqr_mean] = 0.
	win_sqr_mean[win_sqr_mean<=0.] = 0. # add this to prevent sqrt from breaking
	img = np.sqrt(win_sqr_mean - win_mean**2)

	return img

def laplace_filter(img, output=float, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))
	return snd.laplace(img.astype(float), *kwargs)

def dog_filter(img, sigma_low, sigma_high, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))
	return difference_of_gaussians(img.astype(float), sigma_low, sigma_high, *kwargs)

def otsu_filter(img, *kwargs):
	thresh = threshold_otsu(img.astype(float))
	binary = img >= thresh
	return binary.astype(float)

def local_filter(img, *kwargs):
	thresh = threshold_local(img.astype(float), *kwargs)
	binary = img >= thresh
	return binary.astype(float)

def niblack_filter(img, *kwargs):
	thresh = threshold_niblack(img, *kwargs)
	binary = img >= thresh
	return binary.astype(float)

def sauvola_filter(img, *kwargs):
	thresh = threshold_sauvola(img, *kwargs)
	binary = img >= thresh
	return binary.astype(float)

def log_filter(img, sigma, *kwargs):
	return snd.gaussian_laplace(img.astype(float), sigma, *kwargs)

def tophat_filter(img, size, connectivity=4, interpolate=True, *kwargs):
	if interpolate:
		img = interpolate_nan(img.astype(float))
	structure = snd.generate_binary_structure(rank=2, connectivity=connectivity)
	img = snd.white_tophat(img.astype(float), structure=structure, size=size, *kwargs)
	return img
